"""
Evaluate the polar maps pixel-by-pixel.
"""
from typing import List, Optional

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

plt.rcParams.update({"text.usetex": True, "font.size": 18})

# see https://matplotlib.org/stable/gallery/lines_bars_and_markers/linestyles.html
linestyle_loosely_dashed = (0, (5, 10))


def normaliza_polar_map(polar_map: np.ndarray) -> np.ndarray:
    """
    Normalize a polar map loaded as a grayscale image to a range of [0, 100].

    Parameters
    ----------
    polar_map: np.ndarray
        grayscale image of a polar map

    Returns
    -------
    np.ndarray
        polar map image scaled to [0, 100]
    """
    return polar_map * (100.0 / 255.0)


def plot_hist2d(
    data1: np.ndarray,
    data2: np.ndarray,
    label1: str,
    label2: str,
    bins: List[float],
    ax: Optional[mpl.axes.Axes] = None,
):
    """
    Plot a 2D histogram of two data arrays to visualize their correlation.

    Parameters
    ----------
    data1: np.ndarray
        first data array
    data2: np.ndarray
        second data array
    label1: str
        label for the first data array
    label2: str
        label for the second data array
    bins: list of float
        list to describe the binning for the histogram
    ax: mpl.axes.Axes
        axes to draw on - if none a new one is created
    """
    ax = plt.subplot() if ax is None else ax
    v_min = min(bins)
    v_max = max(bins)

    ax.hist2d(data1, data2, bins=bins)
    ax.plot(
        (v_min, v_max),
        (v_min, v_max),
        color="red",
        linestyle=linestyle_loosely_dashed,
    )
    ax.set_xlabel(label1)
    ax.set_ylabel(label2)


if __name__ == "__main__":
    import argparse
    import os

    import cv2 as cv
    import pandas as pd
    from scipy.stats import pearsonr

    from mu_map.polar_map.prepare import headers
    from mu_map.polar_map.visualize import get_circular_mask

    parser = argparse.ArgumentParser(
        description="Compare polar maps pixel by pixel",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--polar_map_dir",
        type=str,
        default="data/polar_maps",
        help="directory containing the polar map images",
    )
    parser.add_argument(
        "--images_dir",
        type=str,
        default="images",
        help="sub-directory under <polar_map_dir> containing the actual image files",
    )
    parser.add_argument(
        "--csv",
        type=str,
        default="polar_maps.csv",
        help="file under <polar_map_dir> containing meta information about the polar maps",
    )
    parser.add_argument(
        "--baseline",
        choices=["symbia", "stir"],
        default="stir",
        help="select the polar map treated as the baseline",
    )
    parser.add_argument(
        "--bin_width",
        type=float,
        default=2,
        help="the bin width used for the 2D histogram",
    )
    parser.add_argument(
        "--save",
        type=str,
        help="save the resulting plot",
    )
    args = parser.parse_args()

    args.images_dir = os.path.join(args.polar_map_dir, args.images_dir)
    args.csv = os.path.join(args.polar_map_dir, args.csv)
    args.baseline = "ct" if args.baseline == "stir" else "symbia"

    pixels_ctac = np.array([])
    pixels_dlac = np.array([])
    pixels_noac = np.array([])

    meta = pd.read_csv(args.csv)
    for _id in meta[headers.id].unique():
        print(f"Load {_id:03d} ...", end="\r")
        _meta = meta[(meta[headers.id] == _id) & ~(meta[headers.segments])]

        file_pm_ctac = _meta[
            (_meta[headers.type] == args.baseline) & _meta[headers.ac]
        ][headers.file].values[0]
        file_pm_dlac = _meta[_meta[headers.type] == "dl"][headers.file].values[0]
        file_pm_noac = _meta[~_meta[headers.ac]][headers.file].values[0]

        pm_ctac = normaliza_polar_map(
            cv.imread(os.path.join(args.images_dir, file_pm_ctac), cv.IMREAD_GRAYSCALE)
        )
        pm_dlac = normaliza_polar_map(
            cv.imread(os.path.join(args.images_dir, file_pm_dlac), cv.IMREAD_GRAYSCALE)
        )
        pm_noac = normaliza_polar_map(
            cv.imread(os.path.join(args.images_dir, file_pm_noac), cv.IMREAD_GRAYSCALE)
        )

        mask = get_circular_mask(pm_ctac.shape)

        pixels_ctac = np.concatenate((pixels_ctac, pm_ctac[mask]))
        pixels_dlac = np.concatenate((pixels_dlac, pm_dlac[mask]))
        pixels_noac = np.concatenate((pixels_noac, pm_noac[mask]))
    print(" " * 100, end="\r")

    def print_stats(data, label):
        print(f"{label}:")

        r, p = pearsonr(data, pixels_ctac)
        dist_abs = np.absolute(data - pixels_ctac)
        dist_per = 100 * dist_abs / pixels_ctac

        print(f" -         Correlation: R={r:.5f}, (p={p:.5f})")
        print(f" -   Absolute Distance: {dist_abs.mean():.5f}±{dist_abs.std():.5f}")
        print(f" - Percentage Distance: {dist_per.mean():.5f}±{dist_per.std():.5f}")

    print_stats(pixels_dlac, "DLAC")
    print_stats(pixels_noac, "No AC")

    bins = np.arange(0, 100 + 0.95 * args.bin_width, args.bin_width)
    fig, axs = plt.subplots(1, 2, figsize=(10, 4.5))

    plot_hist2d(pixels_ctac, pixels_dlac, "CTAC", "DLAC", bins, ax=axs[0])
    plot_hist2d(pixels_ctac, pixels_noac, "CTAC", "No AC", bins, ax=axs[1])

    plt.tight_layout()
    if args.save:
        plt.savefig(args.save, dpi=300)
    plt.show()
