"""
Evaluate the training time for random search training.

Note that evaluation became more difficult because the log formatting
did not use 24h time and does not indicate AM or PM.
"""
import argparse
import os
import datetime

import numpy as np

from mu_map.random_search.eval.util import load_data
from mu_map.logging import parse_file

parser = argparse.ArgumentParser(
    description="Evalute the time a random search procedure took",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)
parser.add_argument(
    "--random_search_dir",
    type=str,
    default="cgan_random_search/",
    help="the directory containing the random search data",
)
args = parser.parse_args()

data = load_data(args.random_search_dir)

timedeltas = []
for run in data.keys():
    log = parse_file(
        os.path.join(args.random_search_dir, data[run]["dir"], "train.log")
    )
    time_start = log[0].time
    time_end = log[-1].time

    if time_start.day != time_end.day:
        """
        Compensate for cases where the day changed during training.
        This means the start time is PM and has to be increased by 12 hours.
        """
        time_start = time_start + datetime.timedelta(hours=12)

    timedelta = time_end - time_start

    if timedelta.total_seconds() < 0:
        """
        Compensate for cases where the training went from AM to PM.
        In this case the time delta is negative.
        """
        time_end = log[-1].time
        time_end = time_end + datetime.timedelta(hours=12)
        timedelta = time_end - log[0].time

    timedeltas.append(timedelta)

seconds = list(map(lambda td: td.total_seconds(), timedeltas))
seconds = np.array(seconds).astype(int)


def format_seconds(seconds: float, with_days=False) -> str:
    seconds = round(seconds)

    if with_days:
        days, seconds = divmod(seconds, 3600 * 24)

    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)

    if with_days:
        return f"{days} {hours:02d}:{minutes:02d}:{seconds:02d}"
    return f"{hours:02d}:{minutes:02d}:{seconds:02d}"


print("   Min:", format_seconds(seconds.min()))
print("   Max:", format_seconds(seconds.max()))
print("Median:", format_seconds(int(np.median(seconds))))
print("  Mean:", format_seconds(seconds.mean()))
print("   Std:", format_seconds(seconds.std()))
print(" Total:", format_seconds(seconds.sum(), with_days=True))
