import random
from typing import Any, Callable, Dict, List


class ParamSampler:
    """
    Abstract class to sample a parameter.
    """

    def sample(self) -> Any:
        """
        Create a new value for a parameter.
        """
        pass


class ChoiceSampler(ParamSampler):
    """
    Sample from a list of choices.
    """

    def __init__(self, values: List[Any]):
        """
        Create a new choice sampler.

        :param values: the list of values from which a sample is drawn.
        """
        super().__init__()
        self.values = values

    def sample(self) -> Any:
        """
        Retrieve a random value from the list of choices.
        """
        idx = random.randrange(0, len(self.values))
        return self.values[idx]


class DependentChoiceSampler(ChoiceSampler):
    """
    A choice sampler that depends on other parameters.
    """

    def __init__(self, build_choices: Callable[[Any], List[Any]]):
        """
        Create a dependent choice sampler.

        :param build_choices: a callable to create a list of choices depending on other parameters
        """
        super().__init__(values=[])

        self.build_choices = build_choices
        self.dependency_names = list(build_choices.__annotations__.keys())

    def sample(self, dependencies: Dict[str, Any]) -> List[Any]:
        """
        Sample a choice based on given dependencies.

        :param dependencies: a dict of name value pairs for all dependencies
                             note that the name has to match the parameter name in the build_choices callable
        :return: a new sample
        """
        self.validate_deps(dependencies)
        self.values = self.build_choices(**dependencies)
        return super().sample()

    def validate_deps(self, dependencies: Dict[str, Any]):
        """
        Validate a dict of dependencies by checking if all required parameters of
        the build_choices callable are available.

        :param dependencies: dict of name value pairs for dependencies
        """
        for name in self.dependency_names:
            assert (
                name in dependencies.keys()
            ), f"Dependency {name} is missing from provided dependencies {dependencies}"


class FloatIntervalSampler(ParamSampler):
    """
    Sample a value from a float interval.
    """

    def __init__(self, min_val: float, max_val: float):
        """
        Create a new float interval sampler.

        :param min_val: the minimal value to draw
        :param max_val: the maximal value to draw
        """
        super().__init__()
        self.min_val = min_val
        self.max_val = max_val

    def sample(self) -> float:
        """
        Draw a new float.
        """
        return random.uniform(self.min_val, self.max_val)


class IntIntervalSampler(ParamSampler):
    """
    Sample a value from an integer interval.
    """

    def __init__(self, min_val: int, max_val: int):
        """
        Create a new int interval sampler.

        :param min_val: the minimal value to draw
        :param max_val: the maximal value to draw
        """
        super().__init__()
        self.min_val = min_val
        self.max_val = max_val

    def sample(self) -> int:
        """
        Draw a new int.
        """
        return random.randint(self.min_val, self.max_val)
