import argparse
import os

import numpy as np
import pydicom
import torch

from mu_map.data.prepare import headers
from mu_map.data.remove_bed import add_bed
from mu_map.dataset.default import MuMapDataset
from mu_map.dataset.normalization import norm_choices, norm_by_str
from mu_map.dataset.transform import SequenceTransform, PadCropTranform
from mu_map.file.dicom import load_dcm, update_dcm, change_uid
from mu_map.logging import add_logging_args, get_logger_by_args
from mu_map.models.unet import UNet

parser = argparse.ArgumentParser(
    description="Compute mu maps predicted by a model",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)
parser.add_argument(
    "--dataset_dir", type=str, default="data/second", help="directory of the dataset"
)
parser.add_argument(
    "--out_dir", type=str, required=True, help="directory where results are stored"
)
parser.add_argument(
    "--device",
    type=str,
    default="cuda" if torch.cuda.is_available() else "cpu",
    choices=["cpu", "cuda"],
    help="the device on which the model is evaluated (cpu or cuda)",
)

parser.add_argument("--weights", type=str, required=True, help="weights of the model")
parser.add_argument(
    "--features",
    type=int,
    nargs="+",
    default=[64, 128, 256, 512],
    help="features defining the U-Net model",
)
parser.add_argument(
    "--split",
    type=str,
    default="test",
    choices=["train", "test", "validation", "all"],
    help="the split of the dataset to be processed",
)
parser.add_argument(
    "--norm",
    type=str,
    choices=norm_choices,
    default="mean",
    help="type of normalization applied to the reconstructions",
)
parser.add_argument(
    "--size",
    type=int,
    default=32,
    help="pad/crop the third tensor dimension to this value",
)
parser.add_argument(
    "--params",
    type=str,
    help="define a parameter definition of a random search iteration to load all parameters from that",
)
add_logging_args(parser, defaults={"--logfile": "compute_mu_maps.log"})

args = parser.parse_args()
if args.params:
    from mu_map.random_search.cgan import load_params

    params = load_params(args.params)

    args.norm = params["normalization"]
    args.features = params["generator_features"]
else:
    args.norm = norm_by_str(args.norm)
args.logfile = os.path.join(args.out_dir, args.logfile)

logger = get_logger_by_args(args)
logger.info(args)

torch.set_grad_enabled(False)
device = torch.device(args.device)

model = UNet(features=args.features)
model.load_state_dict(torch.load(args.weights, map_location=device))
model = model.to(device).eval()

transform_normalization = SequenceTransform(
    [args.norm, PadCropTranform(dim=3, size=32)]
)
dataset = MuMapDataset(
    args.dataset_dir,
    transform_normalization=transform_normalization,
    split_name=args.split,
)
dataset_with_bed = dataset.copy(dataset.split_name, bed_contours_file=None)

for i, ((recon_nac, _), (_, mu_map)) in enumerate(zip(dataset, dataset_with_bed)):
    row = dataset.table.iloc[i]
    _id = row[headers.id]
    logger.debug(f"Process Id: {_id}")

    logger.debug("Predict mu map")
    mu_map = mu_map.squeeze().cpu().numpy()
    recon_nac = recon_nac.to(device)

    logger.debug("Post-process mu map")
    mu_map_dl = model(recon_nac.unsqueeze(dim=0))
    mu_map_dl = mu_map_dl.squeeze().cpu().numpy()
    mu_map_dl = np.clip(mu_map_dl, 0, mu_map_dl.max())
    mu_map_dl = add_bed(mu_map_dl, mu_map, dataset.bed_contours[_id])

    file_mu_map = os.path.join(dataset.dir_images, row[headers.file_mu_map])
    logger.info(f"Load reference DICOM file: {file_mu_map}")
    dcm_mu_map, _ = load_dcm(file_mu_map, direction=1)

    dcm_mu_map_dl = update_dcm(dcm_mu_map.copy(), mu_map_dl)
    dcm_mu_map_dl = change_uid(dcm_mu_map_dl)

    base, ext = os.path.splitext(os.path.basename(file_mu_map))
    file_mu_map_dl = os.path.join(args.out_dir, f"{base}_dl{ext}")
    logger.info(f"Write result to: {file_mu_map_dl}")
    pydicom.dcmwrite(file_mu_map_dl, dcm_mu_map_dl)
